///////////////////////////////////////////////////////////////////////////////////////
//
// Net radio tuner system
//   
//   Radio tuner consists of ONKYO T-422M, Pi Zero, PIC18F, LAN board and DAC board. 
//   This software controls Pi Zero.
//
//   Briefing document(in Japanese) is available on https://www.snowinnovember.info   
//
///////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <termios.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>


//Defines
#include "lm7001_t-422m.h"
#include "command.h"

 //GPIO
 //Pi3
 //#define PERI_BASE  0x3F000000
 //Pi Zero
#define PERI_BASE  0x20000000

#define GPIO_BASE  (PERI_BASE + 0x200000) /* GPIO controller */
#define BLOCK_SIZE (1024 * 4)

#define GPIO_GPFSEL0 (0x0000/4)
#define GPIO_GPFSEL1 (0x0004/4)

#define GPIO_FSEL_INPUT   0x0
#define GPIO_FSEL_OUTPUT  0x1
#define GPIO_FSEL_ALT0    0x4
#define GPIO_FSEL_ALT3    0x7

//Command buffer for PIC to Pi
#define PIC2Pi_CMD_BUF_LEN (128 * 3)

//Radio player configuration 
#define RADIO_PLAYER_CSV        "radio_player.csv"
#define NUM_OF_RADIO_PLAYER     20
#define SCRIPT_FILE_NAME_LEN    (63 + 1)
#define PLAYER_SCRIPT_ITEMS_MAX (10 + 1)
#define PLAYER_SCRIPT_BUF_LEN   512

//Station configuration
#define STATION_CSV           "station.csv"
#define STATION_CSV_LINE_LEN  384
#define STATION_NAME_LEN      (63 + 1)
#define STATION_URL_LEN       (255 + 1)


//Ping command to check internet connection
#define PING_CMD "ping -c1 www.google.co.jp" 

//UART 
#define SERIAL_PORT "/dev/serial0"


//Information of each station
//channel, freq, pll_d and pll_s are redundant information
typedef struct{
  int   channel;
  int   freq;
  int   pll_d;
  int   pll_s;
  char  *name;
  int   player;
  char  *url;
  int   valid;
} Station;

Station station_error = {0, 0, 0, 0, NULL, 0, NULL, 0};


//GPIO access
volatile unsigned *gpio_reg;
static   int      mem_fd = 0;

//Function prototype
volatile unsigned * io_mapping(int);
int  regist_radio_player(char *[]);
int  list_radio_player(char *[]);
int  regist_station(Station *, Station *, char *[]);
void list_station(Station *, Station *, char *[]);
Station *get_cur_station(Station *, Station *, int  , int );
void print_station(Station *);
void regist_valid_for_pic(Station *, Station *, unsigned char *, unsigned char *);


/////////////////////////////////////////////////////////////
//
// Main function 
//
/////////////////////////////////////////////////////////////

void main() {

  int i, j, k;

//UART
  int   fd_sci;
  struct termios uart_options;

  char  pic_data;
  char  pic_data_next;
  char  pic_data_buf[PIC2Pi_CMD_BUF_LEN];
  char  *pic_data_buf_ptr;
  int   pic_data_buf_num;

//Pi2PIC command
  char  pic_send_command;

//For main process
  int   latest_pll_d = 0;
  int   latest_pll_s = 0;
  int   restart_player = 0;

  char  *radio_player_p[NUM_OF_RADIO_PLAYER];
  char  player_script[PLAYER_SCRIPT_BUF_LEN];
  char  *player_script_p[PLAYER_SCRIPT_ITEMS_MAX];

  Station stations_am[NUM_OF_AM_CHN];
  Station stations_fm[NUM_OF_FM_CHN];
  Station *station_cur = &station_error;

  unsigned char  fm_valid[NUM_OF_BYTES_FOR_FM]; //1 bit for one station
  unsigned char  am_valid[NUM_OF_BYTES_FOR_AM];
  unsigned short chksum;

  //For 8-7encoding
  unsigned char  fm_valid7[bytes_8_7encoded(NUM_OF_BYTES_FOR_FM)]; 
  unsigned char  am_valid7[bytes_8_7encoded(NUM_OF_BYTES_FOR_AM)]; 

//For child process
  pid_t child_pid = -1;
  pid_t pid;
  int   child_status;


//GPIO configuration
  gpio_reg = io_mapping(GPIO_BASE);

  //GPIO5
  *( gpio_reg + GPIO_GPFSEL0 ) = (*( gpio_reg + GPIO_GPFSEL0 ) & ~(0x7 << 15)) | (GPIO_FSEL_OUTPUT << 15);

  //GPIO14
  *( gpio_reg + GPIO_GPFSEL1 ) = (*( gpio_reg + GPIO_GPFSEL1 ) & ~(0x7 << 12)) | (GPIO_FSEL_ALT0 << 12);

  //GPIO15
  *( gpio_reg + GPIO_GPFSEL1 ) = (*( gpio_reg + GPIO_GPFSEL1 ) & ~(0x7 << 15)) | (GPIO_FSEL_ALT0 << 15);


//Radio player registration
  if(regist_radio_player(radio_player_p)) {
    exit(1);
  }

  list_radio_player(radio_player_p);
    

//Valid frequency registration
  if(regist_station(stations_fm, stations_am, radio_player_p)) {
    exit(1);
  }

  list_station(stations_fm, stations_am, radio_player_p);

//Valid table for PIC registration
  regist_valid_for_pic(stations_fm, stations_am, fm_valid, am_valid);


//UART configuration
  fd_sci = open(SERIAL_PORT,  O_RDWR | O_NOCTTY| O_NDELAY); // | O_NDELAY|O_NONBLOCK);

  if (fd_sci == -1) {
    perror("Can't open sci\n");
    exit(1);
  }


  tcgetattr(fd_sci, &uart_options);
  uart_options.c_cflag = B115200 | CS8 | CLOCAL | CREAD;
  uart_options.c_iflag = IGNPAR | ICRNL;
  uart_options.c_oflag = 0;
  uart_options.c_lflag = 0;

  cfmakeraw(&uart_options); //To use Binary
  tcflush(fd_sci, TCIFLUSH);

  tcsetattr(fd_sci, TCSANOW, &uart_options);

//make chksum for fm_valid
  chksum = 0;
  for (i = 0; i < NUM_OF_BYTES_FOR_FM - 2; i++) {
    chksum += fm_valid[i];  
  }
  fm_valid[i++] = chksum >> 8;
  fm_valid[i] = chksum & 0xff;  


//make chksum for am_valid
  chksum = 0;
  for (i = 0; i < NUM_OF_BYTES_FOR_AM - 2; i++) {
    chksum += am_valid[i];  
  }
  am_valid[i++] = chksum >> 8;
  am_valid[i] = chksum & 0xff;  


//8-7encode for fm_valid
  for (i = j = 0; i < NUM_OF_BYTES_FOR_FM; i++) {
    if ((i % 7) == 0) {
      k = j;
      j++; 
      fm_valid7[k] = 0;
    }
    fm_valid7[j] = fm_valid[i] & 0x7f;
    fm_valid7[k] = (fm_valid7[k] >> 1) | ((fm_valid[i] & 0x80) >> 1);
    j++; 
  }
  fm_valid7[k] = fm_valid7[k] >> (6 - (--i % 7)); 


//8-7encode for am_valid
  for (i = j = 0; i < NUM_OF_BYTES_FOR_AM; i++) {
    if ((i % 7) == 0) {
      k = j;
      j++; 
      am_valid7[k] = 0;
    }
    am_valid7[j] = am_valid[i] & 0x7f;
    am_valid7[k] = (am_valid7[k] >> 1) | ((am_valid[i] & 0x80) >> 1);
    j++; 
  }
  am_valid7[k] = am_valid7[k] >> (6 - (--i % 7)); 

//Waiting for ethernet connection
  while(system( PING_CMD ) != 0) { 
    sleep(1);
  } 

//Inform boot-up to PIC
  pic_send_command = Pi2PIC_Pi_BOOTUP;
  if(write(fd_sci, &pic_send_command, 1) == -1){
    puts("Error on sending boot-up  to PIC");
    exit(1);
  }
 

//Here is main process loop
  pic_data_buf_num = 0;

  while(1) {
    if (pic_data_buf_num == 0) {

      pic_data_buf_num = read(fd_sci, pic_data_buf, sizeof(pic_data_buf));

      if (pic_data_buf_num < 0) {
        pic_data_buf_num = 0;
      }
        
      if (pic_data_buf_num > 0) {
        pic_data_buf_ptr = pic_data_buf;
      }
      else {
        usleep(50); //to reduce CPU load
        continue; //no data in pic_data_buf
      } 
    }

    
    if (pic_data_buf_num != 0) {
      pic_data = *pic_data_buf_ptr;
      pic_data_buf_ptr++;
      pic_data_buf_num--;

  //Skip PIC2Pi commands which do not have to be executed
      while(pic_data_buf_num >= 5) { //following command is also pll freq? 
        pic_data_next = *(pic_data_buf_ptr + 2);
        if (pic_data_next == PIC2Pi_FM_PLL ||  pic_data_next == PIC2Pi_AM_PLL) {
          pic_data = pic_data_next;
          pic_data_buf_ptr += 3;
          pic_data_buf_num -= 3;
        }
        else {
          break;
        }
      }

        switch ( pic_data ) {
          case PIC2Pi_FM_PLL:
                    if (pic_data_buf_num < 2 ) {
                      puts("\nError on recieve FM pll_d from PIC");
                      exit(1);
                    }
                    latest_pll_d = *pic_data_buf_ptr + (*(pic_data_buf_ptr + 1) << 7);
                    pic_data_buf_ptr += 2;
                    pic_data_buf_num -= 2;

                    latest_pll_s = 1; //1: FM
                    break;

          case PIC2Pi_AM_PLL:
                    if (pic_data_buf_num < 2 ) {
                      puts("\nError on recieve AM pll_d from PIC");
                      exit(1);
                    }
                    latest_pll_d = *pic_data_buf_ptr + (*(pic_data_buf_ptr + 1) << 7);
                    pic_data_buf_ptr += 2;
                    pic_data_buf_num -= 2;

                    latest_pll_s = 0; //0: AM
                    break;

          case PIC2Pi_RESTART_PLAYER:
                    puts("\nRestart player requested");
                    restart_player = 1; //Set flag of restart request
                    break;

          case PIC2Pi_SHUTDOWN:
                    puts("\nShutdown requested");
                    system("sudo shutdown now");
                    break;
                
          case PIC2Pi_CHECK_STATUS:
                    //Send ready to PIC
                    puts("\nStatus check requested");
                    while(system( PING_CMD ) != 0) { //Waiting for ethernet connection
                      sleep(1);
                    } 
                    puts("Send READY to PIC");
                    pic_send_command = Pi2PIC_Pi_READY;
                    if(write(fd_sci, &pic_send_command, 1) == -1){
                      puts("\nError on sending ready  to PIC");
                      exit(1);
                    }
                    break;

          case PIC2Pi_REQ_VALID_TBL:
                    //Send FM valid to PIC
                    pic_send_command = Pi2PIC_REGIST_FM_VALID;
                    if(write(fd_sci, &pic_send_command, 1) == -1){
                      puts("Error on sending FM valid data(header) to PIC");
                      exit(1);
                    }

                    for (i = 0; i < bytes_8_7encoded(NUM_OF_BYTES_FOR_FM); i++) {
                      if(write(fd_sci, fm_valid7 + i, 1) == -1){
                        puts("Error on sending FM valid data to PIC");
                        exit(1);
                      }
                    }


                    //Send AM valid to PIC
                    pic_send_command = Pi2PIC_REGIST_AM_VALID;
                    if(write(fd_sci, &pic_send_command, 1) == -1){
                      puts("Error on sending AM valid data(header) to PIC");
                      exit(1);
                    }

                    for (i = 0; i < bytes_8_7encoded(NUM_OF_BYTES_FOR_AM); i++) {
                      if(write(fd_sci, am_valid7 + i, 1) == -1){
                        puts("Error on sending AM valid data to PIC");
                        exit(1);
                      }
                    }


                    //Get current frequency from PIC
                    pic_send_command = Pi2PIC_REQUEST_FREQ;
                    if(write(fd_sci, &pic_send_command, 1) == -1){
                      puts("Error on sending req freq to PIC");
                      exit(1);
                    }

                    break;
                    
        }
      }


    if ((restart_player == 0) && (latest_pll_d == station_cur->pll_d) && (latest_pll_s == station_cur->pll_s)){
    
      continue;

    }

    //Select station and play
    restart_player = 0;//Clear flag of restart request

    if (child_pid != -1) { // Child processes exist
      puts("Stop playing");
      kill(-getpgid(child_pid), SIGTERM);  
      waitpid(child_pid, &child_status, 0); //in case of defunct
    }

    station_cur = get_cur_station(stations_fm, stations_am, latest_pll_d, latest_pll_s);

    print_station(station_cur);

    if (station_cur->valid == 1) { //Selected station is valid

      if((radio_player_p[station_cur->player]) == NULL ){
        puts("Ileegal player number");
        exit(1);
      }
      else {
        strcpy(player_script, radio_player_p[station_cur->player]);
        strcat(player_script, " ");
        strcat(player_script, station_cur->url);
      }

      printf("Player script: %s\n",player_script);

      for (i = 0; i < PLAYER_SCRIPT_ITEMS_MAX; i++) { 
        if (i == 0) {
          player_script_p[i] = strtok(player_script, " \t"); 
        }
        else {
          player_script_p[i] = strtok(NULL, " \t"); 
        }

        if (player_script_p[i] == NULL) {
          break;
        }
      } 
      if (i == PLAYER_SCRIPT_ITEMS_MAX) { //parameters still remain
        puts("Too many player command parameters");
        exit(1);
      }

      pid = fork();

      if (pid < 0) { //making child process for radio player failed
        child_pid = -1;
      }
      else { 
        if (pid == 0) { //Child process
          setpgid(getpid(), 0);// Change child process group id
          execvp(player_script_p[0], player_script_p);

          exit(1); //error occured on exec

        }
        else {//Parent process
          child_pid = pid;
        }
      }  
    }
  }

}



//Functions
volatile unsigned * io_mapping(int base_addr) {

  void *gpio_map;
    
  if (!mem_fd) {
    if((mem_fd = open("/dev/gpiomem", O_RDWR|O_SYNC)) < 0) {
      puts("Can't open /dev/gpiomem, it might be fixed by using sudo command");
    }
  }
    
  gpio_map = mmap(
                    NULL,
                    BLOCK_SIZE, 
                    PROT_READ|PROT_WRITE, 
                    MAP_SHARED, 
                    mem_fd, 
                    base_addr
                    );

  if(gpio_map == MAP_FAILED) {
    puts("mmap error");
  }

//    close(mem_fd);

  return (volatile unsigned *)gpio_map;
}


int regist_radio_player(char *radio_player_p[]) {
  int   number;
  FILE  *fp;
  char  *fname = RADIO_PLAYER_CSV;
  char  script_fname[SCRIPT_FILE_NAME_LEN];
  char  linebuf[STATION_CSV_LINE_LEN];

  fp = fopen( fname, "r" );
  if( fp == NULL ){
    printf("Can't open %s\n", fname );
    return 1;
  }
  //Clear all
  for(number = 0; number < NUM_OF_RADIO_PLAYER; number++){
    radio_player_p[number] = NULL;
  }

  puts("Regist radio players");
  while(fgets( linebuf, sizeof(linebuf), fp) != NULL ){
    if (sscanf(linebuf, "%d,%[^\n]", &number, script_fname) != 2) {
      printf("Format error in %s\n", fname );
      printf("%s\n", linebuf);
      return 1;
    }

    if (number >= NUM_OF_RADIO_PLAYER) {
      puts("Radio player script number is out of range");
      printf("%s\n", linebuf);
      return 1;
    }

    radio_player_p[number] = (char *)malloc(sizeof(script_fname));
    if(radio_player_p[number] == NULL) {
      puts("Memory cannot be allocated!");
      return 1;
    }
    strcpy(radio_player_p[number], script_fname);
  }
  fclose( fp );

  return 0;
}


int list_radio_player(char *radio_player_p[]) {
  int number;

  puts("List of radio player script");
  for(number = 0; number < NUM_OF_RADIO_PLAYER; number++){
    if(radio_player_p[number] != NULL) {
      printf("%2d  %s\n", number, radio_player_p[number]);
    }
  }
}


int regist_station(Station *stations_fm, Station *stations_am, char *radio_player_p[]) {
  int   channel;
  int   freq;
  FILE  *fp;
  char  linebuf[STATION_CSV_LINE_LEN];
  char  *fname = STATION_CSV;
  char  name[STATION_NAME_LEN];
  int   player;
  char  url[STATION_URL_LEN];
  int   valid;
  Station *station_p;
        
  fp = fopen( fname, "r" );
  if( fp == NULL ){
    printf("Can't open %s\n", fname );
    return 1;
  }
  //Clear all
  puts("Clear registered stations");
  for(channel = 0; channel < NUM_OF_FM_CHN; channel++){
    stations_fm[channel].channel = channel;
    stations_fm[channel].freq    = 0;
    stations_fm[channel].pll_d   = fm_chn2pll_d(channel);
    stations_fm[channel].pll_s   = 1;
    stations_fm[channel].name    = NULL;
    stations_fm[channel].player  = 0;
    stations_fm[channel].url     = NULL;
    stations_fm[channel].valid   = 0;
  }
  for(channel = 0; channel < NUM_OF_AM_CHN; channel++){
    stations_am[channel].channel = channel;
    stations_am[channel].freq    = 0;
    stations_am[channel].pll_d   = am_chn2pll_d(channel);
    stations_am[channel].pll_s   = 0;
    stations_am[channel].name    = NULL;
    stations_am[channel].player  = 0;
    stations_am[channel].url     = NULL;
    stations_am[channel].valid   = 0;
  }

  puts("Regist stations");
  while(fgets( linebuf, sizeof(linebuf), fp) != NULL ){
    if(*linebuf == '\n') {
      continue;
    }
    if(sscanf(linebuf, "%d,%[^,],%d,%[^\n]", &freq, name, &player, url ) != 4) { 
      printf("Format error in %s\n", fname );
      printf("Ignored: %s\n", linebuf);
      continue;
    }

    if((freq >= FM_FREQ_MIN) && (freq <= FM_FREQ_MAX)) {
      //FM
      channel = fm_freq2chn(freq); ;
      if(freq != chn2fm_freq(channel)) {
        printf("Illegal freqency in %s\n", fname );
        printf("Ignored: %s\n", linebuf);
        continue;
      }
      station_p = stations_fm;
    }
    else if (!((freq >= AM_FREQ_MIN) && (freq <= AM_FREQ_MAX))) {
      printf("Freqency in %s is out of range\n", fname );
      printf("Ignored: %s\n", linebuf);
      continue;
    }
    else {
      //AM
      channel = am_freq2chn(freq);
      if(freq != chn2am_freq(channel)) {
        printf("Illegal freqency in %s\n", fname );
        printf("Ignored: %s\n", linebuf);
        continue;
      }
      station_p = stations_am;
    }

    if(radio_player_p[player] == NULL) {
      printf("Specified player number %d in %s is not registered\n", player, fname );
      printf("Ignored: %s\n", linebuf);
      continue;
    }
        
    station_p[channel].freq = freq;

    station_p[channel].name = (char *)malloc(sizeof(name));
    if(station_p[channel].name == NULL) {
      puts("Memory cannot be allocated!");
      return 1;
    }
    strcpy(station_p[channel].name, name);

    station_p[channel].player  = player;

    station_p[channel].url = (char *)malloc(sizeof(url));
    if(station_p[channel].url == NULL) {
      puts("Memory cannot be allocated!");
      return 1;
    }
    strcpy(station_p[channel].url, url);

    station_p[channel].valid = 1;
  }
  fclose( fp );

  return 0;
}


void list_station(Station *stations_fm, Station *stations_am, char *radio_player_p[]) {
  int channel;

  puts("Station list of FM");
  puts("Channel Freq PLL_D PLL_S Name Player URL");
  for(channel = 0; channel < NUM_OF_FM_CHN; channel++){
    if (stations_fm[channel].valid == 1) {
      printf( "%3d, %6d, %03x, %d, %s, %s, %s\n", 
      stations_fm[channel].channel,
      stations_fm[channel].freq,
      stations_fm[channel].pll_d,
      stations_fm[channel].pll_s,
      stations_fm[channel].name,
      radio_player_p[stations_fm[channel].player],
      stations_fm[channel].url
      );
    }
  }
  puts("\nStation list of AM");
  puts("Channel Freq PLL_D PLL_S Name Player URL");
  for(channel = 0; channel < NUM_OF_AM_CHN; channel++){
    if (stations_am[channel].valid == 1) {
      printf( "%3d, %4d, %03x, %d, %s, %s, %s\n", 
      stations_am[channel].channel,
      stations_am[channel].freq,
      stations_am[channel].pll_d,
      stations_am[channel].pll_s,
      stations_am[channel].name,
      radio_player_p[stations_am[channel].player],
      stations_am[channel].url
      );
    }
  }
  puts("\n");
}


Station *get_cur_station(Station *stations_fm, Station *stations_am, int  pll_d, int pll_s) {
  Station   *station_cur;
  int  channel;

  if(pll_s  == 1){ // pll_s 1:FM, 0:AM 
    channel = pll_d2fm_chn(pll_d);
    if(channel < NUM_OF_FM_CHN) {
        station_cur = stations_fm + channel;
    }
    else{
      puts("Pll_dx is out of range!");
      station_cur = &station_error;
    }
  }
  else{
    channel = pll_d2am_chn(pll_d);
    if(channel < NUM_OF_AM_CHN) {
        station_cur = stations_am + channel;
    }
    else{
      puts("Pll_dx is out of range!");
      station_cur = &station_error;
    }
  }
  return station_cur;
}


void print_station(Station *station_cur) {
  
  printf( "Current station: %d, %d, %03x, %d, %s, %d, %s, %d\n", 
    station_cur->channel,
    station_cur->freq,
    station_cur->pll_d,
    station_cur->pll_s,
    station_cur->valid ? station_cur->name : "not registered",
    station_cur->player,
    station_cur->valid ? station_cur->url : "not registered",
    station_cur->valid
  );

  return;
}


void regist_valid_for_pic(Station *stations_fm, Station *stations_am, unsigned char *fm_valid, unsigned char *am_valid) {

  int i;

  //Clear valid flags
  for (i = 0; i < NUM_OF_BYTES_FOR_FM; i++) {
    fm_valid[i] = 0;
  }
  for (i = 0; i < NUM_OF_BYTES_FOR_AM; i++) {
    am_valid[i] = 0;
  }

  //Regist valid flags
  for (i = 0; i < NUM_OF_FM_CHN; i++) {
    fm_valid[i / 8] |= stations_fm[i].valid << (i % 8);
  }
  for (i = 0; i < NUM_OF_AM_CHN; i++) {
    am_valid[i / 8] |= stations_am[i].valid << (i % 8);
  }
}



/* End of code */
